#include "descriptors.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fstream>
#include <iostream>
#include <time.h>

using namespace std;

int main( int argc, char** argv )
{
	char *imageFile, *detectorFile;
	string s;
	string outFile;
	int m;
	bool isOut=false;
	
	// Check number of arguments
	if(argc < 2)
	{
		fprintf(stderr, "Usage: ./colorDescriptorCVC <imageFile> <detectorFile> [options]\n");
		return -1;
	}
	imageFile = argv[1];
	detectorFile = argv[2];
	
	// Default Optional Arguments
	int descriptor = 3;
	int patch_size = 21;
	int scalef = 5.0;
	int num_bins = 36;
	double sigma = 21/4.0;
	int smooth_flag = 0;
	bool fast = false;
	int probs = 25;
	
	// Check optional parameters
	while ((m = getopt(argc, argv, "hd:p:s:m:B:S:FP:o:")) != -1)
	{
		switch (m)
		{
			// HELP
			case 'h':
				printf("\nUsage : ./colorDescriptorCVC <imageFile> <detectorFile> [options]\n");
				printf("\n  Options:\n");
				printf("\t-h\tHelp\n\n");
				printf("\t-d\t<descriptor> (default: dcd)\n");
					printf("\t\tColor Naming Descriptor: cn\n");
					printf("\t\tDiscriminative Color Descriptor: dcd\n");
					printf("\t\tJOSA Color Naming: josa\n");
					printf("\t\tHue Descriptor: hue\n");
					printf("\t\tOpponent Descriptor: opp\n");
				printf("\n");
				printf("\t-p\t<int> Patch Size (default: 21)\n");
				printf("\t-s\t<double> Gaussian Sigma (default: 5.0)\n");
				printf("\t-m\t<double> Magnification Factor (default: 5.0)\n");
				printf("\t-o\t<outputFile>\n");
				printf("\n");
				printf("\t-B\t<int> Number of Bins (only for hue/opp, default: 36)\n");
				printf("\t-S\t<int> Smooth Factor (only for hue/opp, default: 1)\n");
				printf("\t-F\t<int> Use fast method (only for dcd)\n");
				printf("\t-P\t<int> Number of Probabilities (11, 25 or 50) (only for dcd, default: 25)\n");
				printf("\n");
				
				return -1;
			// DESCRIPTOR
			case 'd':
				s = optarg;
				if((s == "cn") or (s == "CN"))
				{
					descriptor = 1;
				}
				else
				{
					if((s == "hue") or (s == "HUE"))
					{
						descriptor = 2;
					}
					else
					{
						if((s == "dcd") or (s == "DCD"))
						{
							descriptor = 3;
						}
						else
						{
							if((s == "josa") or (s == "JOSA"))
							{
								descriptor = 4;
							}
							else
							{
								if((s == "opp") or (s == "OPP"))
								{
									descriptor = 5;
								}
								else
								{
									printf("Unknown descriptor: %s\n", optarg);
									return -1;
								}
							}
						}
					}
				}
				break;
			// PATCH SIZE
			case 'p':
				patch_size = atoi(optarg);
				if (patch_size <= 0)
				{
					printf("Patch size should be positive\n");
					return -1;
				}
				if (patch_size%2 == 0)
				{
					patch_size++;
				}
				break;
			// SIGMA
			case 's':
				sigma = atof(optarg);
				if (sigma < 0.0)
				{
					printf("Sigma should be positive\n");
					return -1;
				}
				break;
			// SCALE FACTOR
			case 'm':
				scalef = atof(optarg);
				if (scalef < 0.0)
				{
					printf("Magnification factor should be positive\n");
					return -1;
				}
				break;
			// NUMBER OF BINS
			case 'B':
				num_bins = atoi(optarg);
				if (num_bins <= 0)
				{
					printf("Number of bins should be positive\n");
					return -1;
				}
				break;
			// SMOOTH FACTOR
			case 'S':
				smooth_flag = atoi(optarg);
				if (smooth_flag < 0)
				{
					printf("Smooth factor should be positive or zero\n");
					return -1;
				}
				break;
			// FAST DCD
			case 'F':
				fast=true;
				break;
			// NUMBER OF PROBABILITIES
			case 'P':
				probs = atoi(optarg);
				if ((probs!=11) && (probs!=25) && (probs!=50))
				{
					printf("Number of probabilities should be 11, 25 or 50\n");
					return -1;
				}
				break;
			// OUTPUT
			case 'o':
				isOut=true;
				outFile.assign(optarg);
				// TODO: check if file is correct
				printf("Some issues with generating the output stream. Use the terminal command '>' instead.\n");
				return -1;
				break;
			// OTHERWISE
			case '?':
				if ((optopt == 'd') or (optopt == 'p') or (optopt == 's') or (optopt == 'f') or (optopt == 'B') or (optopt == 'S') or (optopt == 'P') or (optopt == 'o'))
				{
					printf("Option -%c requires an argument.\n", optopt);
				}
				else
				{
					if(isprint (optopt))
					{
						printf("Unknown option `-%c'.\n", optopt);
					}
					else
					{
						printf("Unknown option character `\\x%x'.\n", optopt);
					}
				}
				return -1;
			default:
				return -1;
		}
	}
	
	// Change ouput stream
	std::streambuf *buf;
	std::ofstream of;
	if(isOut) {
		of.open(outFile.c_str());
		buf = of.rdbuf();
	} else {
		buf = std::cout.rdbuf();
	}
	std::ostream out(buf);
	
	unsigned t0=clock(),t1;
	
	switch(descriptor) {
		// black, blue, brown, gray, green, orange, pink, purple, red, white, yellow
		case 1: descriptorCN(imageFile,detectorFile,patch_size,scalef,sigma,0,out);
			break;
		case 2: descriptorHUE(imageFile,detectorFile,patch_size,scalef,sigma,num_bins,smooth_flag,out);
			break;
		case 3: descriptorDCD(imageFile,detectorFile,patch_size,scalef,sigma,fast,probs,out);
			break;
		case 4: descriptorCN(imageFile,detectorFile,patch_size,scalef,sigma,1,out);
			break;
		case 5: descriptorOPP(imageFile,detectorFile,patch_size,scalef,sigma,num_bins,smooth_flag,out);
			break;
	}
	
	t1=clock()-t0;
	float t2=((float)t1)/CLOCKS_PER_SEC;
	cout << "Took: " << t2 << "sec" << endl;
	return 0;
}
