#include <mex.h>
#include <math.h>
#include <time.h>

#if defined(_WIN32) || defined(_WIN64)
int round(double x) { 
    return (floor(x + 0.5));
}
#endif

int coord(int row, int col, int numRows) {
	return col * numRows + row;
}
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	const double PI = 3.1415926;
	
	int m,k,n,p,r,cx,cy,x,y,side,w,h,px,py,xi,yi,dx,dy,cont,blue,green,red,hue_bin;
	int imgM,imgN,detCols,detRows,outCols;
	int patch_size,scalef,num_bins,smooth_flag,border,superRows,superCols;
	double scalex,scaley,weight,sum,lumEnergy,sat,hue,sigma;
	double *imgR, *imgG, *imgB, *det, *out, *imgR_blk_ptr, *imgG_blk_ptr, *imgB_blk_ptr, *mask_ptr;
	double *R_ptr, *G_ptr, *B_ptr, *M_ptr, *gau_ptr, *imoutR_ptr, *imoutG_ptr, *imoutB_ptr, *patchMask_ptr, *smooth_ptr;
	mxArray *superImageR, *superImageG, *superImageB, *superMask, *gaussian, *imoutR, *imoutG, *imoutB, *patchMask, *smooth;
	mxArray *imgR_blk, *imgG_blk, *imgB_blk, *mask;
	
	/* Check the inputs*/
	if (nrhs < 4)
		mexErrMsgTxt("Not enough inputs provided.\n");
	
	/* INPUTS ****************** */
    /* Thanks to Andrew Bagdanov for helping fix input issues */
	/* Image pixels (R, G, B) */
	imgR    = mxGetPr(prhs[0]);
	imgG    = mxGetPr(prhs[1]);
	imgB    = mxGetPr(prhs[2]);
	/* Detector */
	det     = mxGetPr(prhs[3]);
	/* Image size */
	imgM    = (int) mxGetM(prhs[0]);
	imgN    = (int) mxGetN(prhs[0]);
	/* Detector size */
	detRows = (int) mxGetM(prhs[3]);
	detCols = (int) mxGetN(prhs[3]);
	/* Default Optional Arguments */
	/* Default patch_size */
	patch_size = 21;
	if (nrhs >= 5) {
	  patch_size  = (int) mxGetScalar(prhs[4]);
	}
	/* Default scalef */
	scalef = 5.0;
	if (nrhs >= 6) {
	  scalef = (int) mxGetScalar(prhs[5]);
	}
	/* Default sigma */
	sigma = ((double) patch_size) / 4.0;
	if (nrhs == 7) {
	  sigma = mxGetScalar(prhs[6]);
	}
	/* Default num_bins */
	num_bins = 18;
	if (nrhs >= 8) {
	  num_bins = (int) mxGetScalar(prhs[7]);
	}
	/* Default smooth_flag */
	smooth_flag = 0;
	if (nrhs >= 9) {
	  smooth_flag = (int) mxGetScalar(prhs[8]);
	}

	if (!mxIsDouble(prhs[0]) | !mxIsDouble(prhs[1]) | !mxIsDouble(prhs[2]))
		mexErrMsgTxt("Input image should be double.\n");
	
	if (!mxIsDouble(prhs[3]))
		mexErrMsgTxt("Detector matrix should be double.\n");
	
	/* OUTPUT ****************** */
	outCols = detCols+num_bins;
	plhs[0] = mxCreateDoubleMatrix(detRows, outCols, mxREAL);
	out = mxGetPr(plhs[0]);
	
	/* Detectors out of image boundaries */
	border = 0;
	for(m=0; m<detRows; m++)
	{
		r = round(det[coord(m,2,detRows)]*scalef);
		if(r>border)
		{
			border = r;
		}
	}
	border = 3*border;
	
	/* Create Super Image for treating detectors out of image bounds */
	superRows = imgM+2*border;
	superCols = imgN+2*border;
	superImageR = mxCreateDoubleMatrix(superRows, superCols, mxREAL);
	superImageG = mxCreateDoubleMatrix(superRows, superCols, mxREAL);
	superImageB = mxCreateDoubleMatrix(superRows, superCols, mxREAL);
	superMask = mxCreateDoubleMatrix(superRows, superCols, mxREAL);
	R_ptr = (double *)mxGetPr(superImageR);
	G_ptr = (double *)mxGetPr(superImageG);
	B_ptr = (double *)mxGetPr(superImageB);
	M_ptr = (double *)mxGetPr(superMask);
	
	for(m=0; m<border; m++)
	{
		for(k=0; k<border; k++)
		{
			/* top-left */
			R_ptr[coord(m,k,superRows)] = imgR[0];
			G_ptr[coord(m,k,superRows)] = imgG[0];
			B_ptr[coord(m,k,superRows)] = imgB[0];
			M_ptr[coord(m,k,superRows)] = 0.;
			/* bottom-left */
			R_ptr[coord(m+border+imgM,k,superRows)] = imgR[coord(imgM-1,0,imgM)];
			G_ptr[coord(m+border+imgM,k,superRows)] = imgG[coord(imgM-1,0,imgM)];
			B_ptr[coord(m+border+imgM,k,superRows)] = imgB[coord(imgM-1,0,imgM)];
			M_ptr[coord(m+border+imgM,k,superRows)] = 0.;
			/* top-right */
			R_ptr[coord(m,k+border+imgN,superRows)] = imgR[coord(0,imgN-1,imgM)];
			G_ptr[coord(m,k+border+imgN,superRows)] = imgG[coord(0,imgN-1,imgM)];
			B_ptr[coord(m,k+border+imgN,superRows)] = imgB[coord(0,imgN-1,imgM)];
			M_ptr[coord(m,k+border+imgN,superRows)] = 0.;
			/* bottom-right */
			R_ptr[coord(m+border+imgM,k+border+imgN,superRows)] = imgR[coord(imgM-1,imgN-1,imgM)];
			G_ptr[coord(m+border+imgM,k+border+imgN,superRows)] = imgG[coord(imgM-1,imgN-1,imgM)];
			B_ptr[coord(m+border+imgM,k+border+imgN,superRows)] = imgB[coord(imgM-1,imgN-1,imgM)];
			M_ptr[coord(m+border+imgM,k+border+imgN,superRows)] = 0.;
		}
	}
	for(m=0; m<imgM; m++)
	{
		for(k=0; k<border; k++)
		{
			/* middle-left */
			R_ptr[coord(m+border,k,superRows)] = imgR[coord(m,0,imgM)];
			G_ptr[coord(m+border,k,superRows)] = imgG[coord(m,0,imgM)];
			B_ptr[coord(m+border,k,superRows)] = imgB[coord(m,0,imgM)];
			M_ptr[coord(m+border,k,superRows)] = 0.;
			/* middle-right */
			R_ptr[coord(m+border,k+border+imgN,superRows)] = imgR[coord(m,imgN-1,imgM)];
			G_ptr[coord(m+border,k+border+imgN,superRows)] = imgG[coord(m,imgN-1,imgM)];
			B_ptr[coord(m+border,k+border+imgN,superRows)] = imgB[coord(m,imgN-1,imgM)];
			M_ptr[coord(m+border,k+border+imgN,superRows)] = 0.;
		}
	}
	for(m=0; m<border; m++)
	{
		for(k=0; k<imgN; k++)
		{
			/* top-center */
			R_ptr[coord(m,k+border,superRows)] = imgR[coord(0,k,imgM)];
			G_ptr[coord(m,k+border,superRows)] = imgG[coord(0,k,imgM)];
			B_ptr[coord(m,k+border,superRows)] = imgB[coord(0,k,imgM)];
			M_ptr[coord(m,k+border,superRows)] = 0.;
			/* bottom-center */
			R_ptr[coord(m+border+imgM,k+border,superRows)] = imgR[coord(imgM-1,k,imgM)];
			G_ptr[coord(m+border+imgM,k+border,superRows)] = imgG[coord(imgM-1,k,imgM)];
			B_ptr[coord(m+border+imgM,k+border,superRows)] = imgB[coord(imgM-1,k,imgM)];
			M_ptr[coord(m+border+imgM,k+border,superRows)] = 0.;
		}
	}
	for(m=0; m<imgM; m++)
	{
		for(k=0; k<imgN; k++)
		{
			/* middle-center */
			R_ptr[coord(m+border,k+border,superRows)] = imgR[coord(m,k,imgM)];
			G_ptr[coord(m+border,k+border,superRows)] = imgG[coord(m,k,imgM)];
			B_ptr[coord(m+border,k+border,superRows)] = imgB[coord(m,k,imgM)];
			M_ptr[coord(m+border,k+border,superRows)] = 1.;
		}
	}
	
	/* Gaussian */
	gaussian = mxCreateDoubleMatrix(patch_size, patch_size, mxREAL);
	gau_ptr = mxGetPr(gaussian);
	for(m=0; m<patch_size; m++)
	{
		for(k=0; k<patch_size; k++)
		{
			if(sigma == 0)
			{
				gau_ptr[coord(m,k,patch_size)] = 1.;
			}
			else
			{
				gau_ptr[coord(m,k,patch_size)] = exp((-(pow(m - ((patch_size - 1)/2), 2) + pow(k - ((patch_size - 1)/2), 2)))/(2*sigma*sigma));
			}
		}
	}
	
	/* Output matrix */
	imoutR = mxCreateDoubleMatrix(patch_size,patch_size, mxREAL);
	imoutG = mxCreateDoubleMatrix(patch_size,patch_size, mxREAL);
	imoutB = mxCreateDoubleMatrix(patch_size,patch_size, mxREAL);
	patchMask = mxCreateDoubleMatrix(patch_size,patch_size, mxREAL);
	imoutR_ptr = mxGetPr(imoutR);
	imoutG_ptr = mxGetPr(imoutG);
	imoutB_ptr = mxGetPr(imoutB);
	patchMask_ptr = mxGetPr(patchMask);
	
	/* Apply to every ROI */
	for(m=0; m<detRows; m++)
	{
		/* Initialize output row */
		for(k=0; k<detCols; k++)
		{
			out[coord(m,k,detRows)] = det[coord(m,k,detRows)];
		}
		for(k=detCols; k<outCols; k++)
		{
			out[coord(m,k,detRows)] = 0.0;
		}
		/* The -1 is because the descriptor comes from MatLab and starts at position 1 instead of 0 */
		cx = det[coord(m,0,detRows)] + border -1;
		cy = det[coord(m,1,detRows)] + border -1;
		r = round(det[coord(m,2,detRows)]*scalef);
		x = cx - r;
		y = cy - r;
		side = (int) 2*r+1;
		scalex  = (double) side / patch_size;
		scaley  = (double) side / patch_size;
		
		/* Check if the roi is not totally inside the image */
		if((x+side>superCols) || (y+side>superRows) || (x<0) || (y<0))
		{
			mexPrintf("[WARN] ROI completely outside of image\n");
			continue;
		}
		
		if(side==patch_size)
		{
			for (k=y; k<y+side; k++)
			{
				for (n=x; n<x+side; n++)
				{
					imoutR_ptr[coord(k-y,n-x,patch_size)] = R_ptr[coord(k,n,superRows)];
					imoutG_ptr[coord(k-y,n-x,patch_size)] = G_ptr[coord(k,n,superRows)];
					imoutB_ptr[coord(k-y,n-x,patch_size)] = B_ptr[coord(k,n,superRows)];
					patchMask_ptr[coord(k-y,n-x,patch_size)] = M_ptr[coord(k,n,superRows)];
				}
			}
		}
		else
		{
			/* Separate the image to resize into 3 channels */
			mxArray *imgR_blk, *imgG_blk, *imgB_blk, *mask;
			double *imgR_blk_ptr, *imgG_blk_ptr, *imgB_blk_ptr, *mask_ptr;
			imgR_blk = mxCreateDoubleMatrix(side, side, mxREAL);
			imgG_blk = mxCreateDoubleMatrix(side, side, mxREAL);
			imgB_blk = mxCreateDoubleMatrix(side, side, mxREAL);
			mask = mxCreateDoubleMatrix(side, side, mxREAL);
			imgR_blk_ptr = mxGetPr(imgR_blk);
			imgG_blk_ptr = mxGetPr(imgG_blk);
			imgB_blk_ptr = mxGetPr(imgB_blk);
			mask_ptr = mxGetPr(mask);
			
			for (k=y; k<y+side; k++)
			{
				for (n=x; n<x+side; n++)
				{
					imgR_blk_ptr[coord(k-y,n-x,side)] = R_ptr[coord(k,n,superRows)];
					imgG_blk_ptr[coord(k-y,n-x,side)] = G_ptr[coord(k,n,superRows)];
					imgB_blk_ptr[coord(k-y,n-x,side)] = B_ptr[coord(k,n,superRows)];
					mask_ptr[coord(k-y,n-x,side)] = M_ptr[coord(k,n,superRows)];
				}
			}
			
			w = side-1;
			h = side-1;
			
			/* Resize roi to patch */
			for(k=0; k<patch_size; k++)
			{
				py = k*scaley;
				yi = (int) py;
				dy = py-yi;
				for(n=0 ; n<patch_size; n++)
				{
					px = n*scalex;
					xi = (int) px;
					dx = px-xi;
					if( (xi<w) && (yi<h) )
					{
						imoutR_ptr[coord(k,n,patch_size)] = (1.-dy) * ((1.-dx)*imgR_blk_ptr[coord(yi,xi,side)] +dx*imgR_blk_ptr[coord(yi,xi+1,side)]) + dy * ((1.-dx)*imgR_blk_ptr[coord(yi+1,xi,side)] +dx*imgR_blk_ptr[coord(yi+1,xi+1,side)]);
						imoutG_ptr[coord(k,n,patch_size)] = (1.-dy) * ((1.-dx)*imgG_blk_ptr[coord(yi,xi,side)] +dx*imgG_blk_ptr[coord(yi,xi+1,side)]) + dy * ((1.-dx)*imgG_blk_ptr[coord(yi+1,xi,side)] +dx*imgG_blk_ptr[coord(yi+1,xi+1,side)]);
						imoutB_ptr[coord(k,n,patch_size)] = (1.-dy) * ((1.-dx)*imgB_blk_ptr[coord(yi,xi,side)] +dx*imgB_blk_ptr[coord(yi,xi+1,side)]) + dy * ((1.-dx)*imgB_blk_ptr[coord(yi+1,xi,side)] +dx*imgB_blk_ptr[coord(yi+1,xi+1,side)]);
						patchMask_ptr[coord(k,n,patch_size)] = (1.-dy) * ((1.-dx)*mask_ptr[coord(yi,xi,side)]  +dx*mask_ptr[coord(yi,xi+1,side)])     + dy * ((1.-dx)*mask_ptr[coord(yi+1,xi,side)]     +dx*mask_ptr[coord(yi+1,xi+1,side)]);
					}
					else if( (xi<=w) && (yi<h) )
					{
						imoutR_ptr[coord(k,n,patch_size)] = (1.-dy) * ((1.-dx)*imgR_blk_ptr[coord(yi,w,side)] +dx*imgR_blk_ptr[coord(yi,w,side)]) + dy * ((1.-dx)*imgR_blk_ptr[coord(yi+1,w,side)] +dx*imgR_blk_ptr[coord(yi+1,w,side)]);
						imoutG_ptr[coord(k,n,patch_size)] = (1.-dy) * ((1.-dx)*imgG_blk_ptr[coord(yi,w,side)] +dx*imgG_blk_ptr[coord(yi,w,side)]) + dy * ((1.-dx)*imgG_blk_ptr[coord(yi+1,w,side)] +dx*imgG_blk_ptr[coord(yi+1,w,side)]);
						imoutB_ptr[coord(k,n,patch_size)] = (1.-dy) * ((1.-dx)*imgB_blk_ptr[coord(yi,w,side)] +dx*imgB_blk_ptr[coord(yi,w,side)]) + dy * ((1.-dx)*imgB_blk_ptr[coord(yi+1,w,side)] +dx*imgB_blk_ptr[coord(yi+1,w,side)]);
						patchMask_ptr[coord(k,n,patch_size)] = (1.-dy) * ((1.-dx)*mask_ptr[coord(yi,w,side)]  +dx*mask_ptr[coord(yi,w,side)])     + dy * ((1.-dx)*mask_ptr[coord(yi+1,w,side)]     +dx*mask_ptr[coord(yi+1,w,side)]);
					}
					else if( (xi<w) && (yi<=h) )
					{
						imoutR_ptr[coord(k,n,patch_size)] = (1.-dy) * ((1.-dx)*imgR_blk_ptr[coord(yi,xi,side)] +dx*imgR_blk_ptr[coord(yi,xi+1,side)]) + dy * ((1.-dx)*imgR_blk_ptr[coord(h,xi,side)] +dx*imgR_blk_ptr[coord(h,xi+1,side)]);
						imoutG_ptr[coord(k,n,patch_size)] = (1.-dy) * ((1.-dx)*imgG_blk_ptr[coord(yi,xi,side)] +dx*imgG_blk_ptr[coord(yi,xi+1,side)]) + dy * ((1.-dx)*imgG_blk_ptr[coord(h,xi,side)] +dx*imgG_blk_ptr[coord(h,xi+1,side)]);
						imoutB_ptr[coord(k,n,patch_size)] = (1.-dy) * ((1.-dx)*imgB_blk_ptr[coord(yi,xi,side)] +dx*imgB_blk_ptr[coord(yi,xi+1,side)]) + dy * ((1.-dx)*imgB_blk_ptr[coord(h,xi,side)] +dx*imgB_blk_ptr[coord(h,xi+1,side)]);
						patchMask_ptr[coord(k,n,patch_size)] = (1.-dy) * ((1.-dx)*mask_ptr[coord(yi,xi,side)]  +dx*mask_ptr[coord(yi,xi+1,side)])     + dy * ((1.-dx)*mask_ptr[coord(h,xi,side)]     +dx*mask_ptr[coord(h,xi+1,side)]);
					}
					else if( (xi<=w) && (yi<=h) )
					{
						imoutR_ptr[coord(k,n,patch_size)] = (1.-dy) * ((1.-dx)*imgR_blk_ptr[coord(yi,w,side)] +dx*imgR_blk_ptr[coord(yi,w,side)]) + dy * ((1.-dx)*imgR_blk_ptr[coord(h,w,side)] +dx*imgR_blk_ptr[coord(h,w,side)]);
						imoutG_ptr[coord(k,n,patch_size)] = (1.-dy) * ((1.-dx)*imgG_blk_ptr[coord(yi,w,side)] +dx*imgG_blk_ptr[coord(yi,w,side)]) + dy * ((1.-dx)*imgG_blk_ptr[coord(h,w,side)] +dx*imgG_blk_ptr[coord(h,w,side)]);
						imoutB_ptr[coord(k,n,patch_size)] = (1.-dy) * ((1.-dx)*imgB_blk_ptr[coord(yi,w,side)] +dx*imgB_blk_ptr[coord(yi,w,side)]) + dy * ((1.-dx)*imgB_blk_ptr[coord(h,w,side)] +dx*imgB_blk_ptr[coord(h,w,side)]);
						patchMask_ptr[coord(k,n,patch_size)] = (1.-dy) * ((1.-dx)*mask_ptr[coord(yi,w,side)]  +dx*mask_ptr[coord(yi,w,side)])     + dy * ((1.-dx)*mask_ptr[coord(h,w,side)]     +dx*mask_ptr[coord(h,w,side)]);
					}
					
					if(patchMask_ptr[coord(k,n,patch_size)]<0.5)
					{
						patchMask_ptr[coord(k,n,patch_size)] = 0.;
					}
					else
					{
						patchMask_ptr[coord(k,n,patch_size)] = 1.;
					}
				}
			}
			mxDestroyArray(imgR_blk);
			mxDestroyArray(imgG_blk);
			mxDestroyArray(imgB_blk);
			mxDestroyArray(mask);
		}
		
		/* Apply HUE to all pixels */
		lumEnergy=0;
		cont=0;
		sum=0;
		for(k=0; k<patch_size; k++)
		{
			for(n=0; n<patch_size; n++)
			{
				if(patchMask_ptr[coord(k,n,patch_size)]==1)
				{
					weight = gau_ptr[coord(k,n,patch_size)];
					blue   = imoutB_ptr[coord(k,n,patch_size)];
					green  = imoutG_ptr[coord(k,n,patch_size)];
					red    = imoutR_ptr[coord(k,n,patch_size)];
					hue = atan2(red+green-2*blue,sqrt(3)*(red-green))+PI;
					hue = (hue/(2*PI))*num_bins;
					hue_bin = floor(hue);
					if(hue_bin >= num_bins)
					{
						hue_bin = 0;
					}
					sat = (2./3.)*(red*red + green*green + blue*blue - red*green - red*blue - green*blue)+0.01;
					lumEnergy = lumEnergy + (red*red + green*green + blue*blue)*weight;
					/* Interpolation */
					/*for(p=hue_bin;p<hue_bin+2;p++) {
						if(p >= num_bins)
						{
							out[coord(m,detCols,detRows)] = out[coord(m,detCols,detRows)] + (1.-fabs(hue-p))*sat*weight;
						}
						else
						{
							out[coord(m,p+detCols,detRows)] = out[coord(m,p+detCols,detRows)] + (1.-fabs(hue-p))*sat*weight;
						}
					}*/
					if(hue_bin >= num_bins)
					{
						out[coord(m,detCols,detRows)] = out[coord(m,detCols,detRows)] + sat*weight;
					}
					else
					{
						out[coord(m,hue_bin+detCols,detRows)] = out[coord(m,hue_bin+detCols,detRows)] + sat*weight;
					}
					cont++;
					sum = sum + weight;
				}
			}
		}
		/* Normalize gaussian */
		/*lumEnergy = lumEnergy/sum;*/
		if(cont>0)
		{
			for(k=detCols; k<outCols; k++)
			{
				out[coord(m,k,detRows)] = sqrt( (out[coord(m,k,detRows)]) /(lumEnergy+0.000001));
				/* If there is any '-nan' */
				if(out[coord(m,k,detRows)] != out[coord(m,k,detRows)])
				{
					out[coord(m,k,detRows)] = 0.0;
				}
			}
		}
		
		/* Smooth - if 3 bins or more */
		if((smooth_flag>0) && (num_bins>2))
		{
			smooth = mxCreateDoubleMatrix(1, num_bins, mxREAL);
			smooth_ptr = mxGetPr(smooth);
			/* Apply Smoothing */
			for(k=0; k<smooth_flag; k++)
			{
				/* Initialize */
				for(n=0; n<num_bins; n++)
				{
					smooth_ptr[n] = 0.0;
				}
				/* first hue_bin */
				smooth_ptr[0] = (2.0*out[coord(m,detCols,detRows)] + out[coord(m,detCols+1,detRows)] + out[coord(m,detCols+num_bins-1,detRows)])/4.0;
				/* last hue_bin */
				smooth_ptr[num_bins-1] = (2.0*out[coord(m,detCols+num_bins-1,detRows)] + out[coord(m,detCols,detRows)] + out[coord(m,detCols+num_bins-2,detRows)])/4.0;
				/* other hue_bins */
				for(n=1; n<(num_bins-1); n++)
				{
					smooth_ptr[n] = (2*out[coord(m,detCols+n,detRows)] + out[coord(m,detCols+n+1,detRows)] + out[coord(m,detCols+n-1,detRows)])/4.0;
				}
				/* Copy to Output */
				for(n=0; n<num_bins; n++)
				{
					out[coord(m,detCols+n,detRows)] = smooth_ptr[n];
				}
			}
			mxDestroyArray(smooth);
		}
	}
	
	mxDestroyArray(superImageR);
	mxDestroyArray(superImageG);
	mxDestroyArray(superImageB);
	mxDestroyArray(superMask);
	mxDestroyArray(gaussian);
	mxDestroyArray(imoutR);
	mxDestroyArray(imoutG);
	mxDestroyArray(imoutB);
	mxDestroyArray(patchMask);
}
